
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	Gerbil
//
//	Copyright (c) 2001, Bruce Moreland.  All rights reserved.
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	This file is part of the Gerbil chess program project.
//
//	Gerbil is free software; you can redistribute it and/or modify it under
//	the terms of the GNU General Public License as published by the Free
//	Software Foundation; either version 2 of the License, or (at your option)
//	any later version.
//
//	Gerbil is distributed in the hope that it will be useful, but WITHOUT ANY
//	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
//	FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
//	details.
//
//	You should have received a copy of the GNU General Public License along
//	with Gerbil; if not, write to the Free Software Foundation, Inc.,
//	59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#include "engine.hpp"
#include "moves.hpp"
#include "eval.hpp"

#ifdef	_DEBUG
static char const s_aszModule[] = __FILE__;
#endif


unsigned int MVV_LVA[16][16];

// Another set of values for the pieces
// Here knight and bishop share the same value because anyway any kind of
// exchange evaluator does not really take care of the position
// We just want to know if we loose/win material or not, we don't care if we have
// changed a bad bishop for a good knight for example or if we made doubled pawns
#define exchangePieceValueMAX 2000

const int exchangePieceValue[16] = {
	2000,
	2000,
	valQUEEN,
	valQUEEN,
	valROOK,
	valROOK,
	valBISHOP,
	valBISHOP,
	valBISHOP,
	valBISHOP,
	valPAWN,
	valPAWN,
	valPAWN,    // EP
	valPAWN,    // EP
    0,
    0
};

//#define cmkVAL(a)   ((a) + valKING)
// this is perhaps not a good idea
//	new (bad idea)		pcm->cmk = cmkCAPTURE | cmkVAL(ppiTo->val - ppiFrom->val);
//	old					pcm->cmk = cmkCAPTURE | cmkVAL(ppiTo->val - ppiFrom->pc);

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	The values in the arrays correspond to the number of squares you have to
//	move in order to go in a particular "chess-useful" direction.

//	For instance, on a 16 x 8 board, 16 is up one rank, and 17 is up one rank
//	and to the right one file.

static int	const s_argvecKnight[] = { 14, 31, 33, 18, -14, -31, -33, -18, 0 };
static int	const s_argvecBishop[] = { 15, 17, -15, -17, 0 };
static int	const s_argvecRook[] = { -1, 1, 16, -16, 0 };
static int	const s_argvecQueen[] = { -1, 1, 16, -16, 15, 17, -15, -17, 0 };

// used in attack.cpp too
int const * s_argpvecPiece[] = {
	NULL,			// Pawn isn't handled through here.
	s_argvecKnight,
	s_argvecBishop,
	s_argvecRook,
	s_argvecQueen,
	s_argvecQueen,	// King and queen use the same ones.
};

//	The first two values are pawn captures, the last is a single-space move.

static int	const s_argvecWhitePawn[] = { 15, 17, 16, };	// <-- Ordering important.
static int	const s_argvecBlackPawn[] = { -15, -17, -16, };	// <-- Ordering important.

// used in attack.cpp too
int const * s_argpvecPawn[] = {
	s_argvecWhitePawn,
	s_argvecBlackPawn,
};

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This stuff helps avoid messy computation when dealing with castling, which
//	is already messy enough.
#if 0
typedef	struct	tagCSTL {
	int	cf;			// Castling flag that's being handled here.
	int	isqRookTo;	// Where the rook goes.
	int	isqKnight;	// Where the knight started out.
	int	isqKingTo;	// Where the king goes.
}	CSTL;

static CSTL const c_argcstl[coMAX][2] = { {
    {cfE1G1,	isqF1,	isqG1,	isqG1},
    {cfE1C1,	isqD1,	isqB1,	isqC1}},
    {{cfE8G8,	isqF8,	isqG8,	isqG8},
    {cfE8C8,	isqD8,	isqB8,	isqC8}}
};

static __forceinline void addMove(PCM &pcm, int _isqFrom, int _isqTo, unsigned _cmk, unsigned _cmf) {
    pcm->isqFrom = _isqFrom;
    pcm->isqTo = _isqTo;
    pcm->cmk = _cmk;
    pcm++->cmf = _cmf;
}
#endif
// Precompute the MVV_LVA array so that I can try some funny things without changing code
// everywhere. Anyway to compute MVV/LVA I would have to do 2 array access, now I only need
// one access
void initMVVLVA(const CON * pcon) {

    for(int capturingPiece = KING; capturingPiece <= EP_CAPT + 1; capturingPiece++)
        for(int capturedPiece = KING; capturedPiece <= EP_CAPT + 1; capturedPiece++) {
            int valCaptured = exchangePieceValue[capturedPiece];
#if 1
            int valCapturing = exchangePieceValue[capturingPiece];
            // scheme 1 the smallest capturing piece first
            // pxQ > pxR > bxQ > bxR > rxB > rxP > qxQ > qxP
            int mvv_lva = valCaptured - valCapturing;
            unsigned int goodCap = (mvv_lva >= 0) ? cmkCAPTURE : cmkBADCAPT;
#else
            // scheme 2 the biggest taken piece first
            // pxQ > bxQ > qxQ > pxR > bxR > rxR > rxP > qxP
            int mvv_lva = valCaptured - (7 - (capturingPiece >> 1) );
            unsigned int goodCap = ((capturedPiece < PAWN) || (capturingPiece == PAWN)) ? cmkCAPTURE : cmkBADCAPT;
#endif
            // store the result, the constant is there to keep numbers positive
            MVV_LVA[capturingPiece][capturedPiece] = goodCap | (mvv_lva + exchangePieceValueMAX);

        }

}
#if 0
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This generates moves and adds them to the move list starting at
//	"pste->pcmFirst".  If "inQuiescent" is set, this performs as a quiescent
//	capture generator.  This could be broken into two functions for speed,
//	but the extra code is not worth it in this simple application.
//
//	Captures are ordered by MVV/LVA, meaning that the sort key is:
//
//		VALUE(captured piece) - INDEX(capturing piece).
//
//	This means that PxK gets the highest value, and KxP gets the lowest.
//
//	MVV/LVA kind of sucks a bit.  If a static-exchange evaluator is written,
//	and pruning is done based upon the result of the static exchange test, the
//	number of nodes processed per ply will drop dramatically, with little or
//	no loss in effective strength per ply.

void VGenMoves(PCON pcon, PSTE pste, bool inQuiescent)
{
	PCM	pcm = pste->pcmFirst;
    int side = pste->coUs;

	Assert(pcm + 500 <			// Dummy check to see if I'm getting close to
		pcon->argcm + ccmMAX);	//  blowing out the move list.

	for (int ipi = 0; ipi < pcon->argcpi[side]; ipi++) {

		PPI ppiFrom = &pcon->argpi[side][ipi];
		if (ppiFrom->fDead)	// Ignore pieces that don't exist anymore.
			continue;
		int isqFrom = ppiFrom->isq;
		switch (ppiFrom->pc) {
			int	i;
			int const * pvec;
			int	isqTo;
			PPI	ppiTo;

		case pcPAWN:
			//
			//	Pawn captures, including en-passant.
			//
			pvec = s_argpvecPawn[ side ];
			for (i = 0; i < 2; i++)		// Handle two captures first.
				if (!((isqTo = isqFrom + *pvec++) & 0x88))
					if ((ppiTo = pcon->argsq[isqTo].ppi) == NULL) {
						//
						//	En-passant case.
						//
						if (pste->isqEnP == isqTo) {
                            addMove( pcm, isqFrom, isqTo, MVV_LVA[pcPAWN][pcPAWN], cmfMAKE_ENP | cmfCAPTURE);
						}
					} else if (ppiTo->co != side)
						if ((isqTo >= isqA8) || (isqTo <= isqH1)) {
							//	Promotion + capturing case.
							//
							for (int pc = pcQUEEN; pc >= pcKNIGHT; pc--) {
                                addMove( pcm, isqFrom, isqTo, MVV_LVA[pcPAWN][pc], cmfCAPTURE | pc);
							}
						} else {
							//
							//	Regular capture.
							//
                            addMove( pcm, isqFrom, isqTo, MVV_LVA[pcPAWN][ppiTo->pc], cmfCAPTURE);
						}
			if (inQuiescent)
				break;
			//
			//	Non-capturing pawn-advance.
			//
			isqTo = isqFrom + *pvec;
			if (pcon->argsq[isqTo].ppi == NULL)
				if ((isqTo >= isqA8) || (isqTo <= isqH1)) {
					//	Non-capturing pawn promotion.
					//
					for (int pc = pcQUEEN; pc >= pcKNIGHT; pc--) {
                        addMove( pcm, isqFrom, isqTo, (pc == pcQUEEN) ? cmkQUEEN : cmkNONE, pc);
					}
				} else {
					//
					//	This is the normal single-step move.
					//
                    addMove( pcm, isqFrom, isqTo, cmkNONE, cmfNONE);
					//
					//	Check for double-step on first move.
					//
					if (((side == coWHITE) && (ppiFrom->isq <= isqH2)) ||
						((side == coBLACK) && (ppiFrom->isq >= isqA7))) {
						isqTo += *pvec;
						if (pcon->argsq[isqTo].ppi == NULL) {
                            addMove( pcm, isqFrom, isqTo, cmkNONE, cmfSET_ENP);
						}
					}
				}
			break;
		case pcKING:
			//
			//	The king has some special castling code, after which it falls
			//	into the knight code.
			//
			//	It's assumed that if a castling flag is on, the king is on its
			//	original square, and there is a rook on the right square, too.
			//
			//	This code checks to see if other appropriate squares are
			//	empty.  It does *not* check to see if the king move puts us
			//	into check.  It *does* check to see if the king move crossed
			//	check, which is kind of expensive, but that's life.  It could
			//	check elsewhere, but you can make *that* change.
			//
			if ((!pste->fInCheck) && (!inQuiescent))
				for (i = 0; i < 2; i++)
					if (pste->cf & c_argcstl[side][i].cf) 
						if ((pcon->argsq[c_argcstl[side][i].isqRookTo].ppi == NULL) &&
							(pcon->argsq[c_argcstl[side][i].isqKnight].ppi == NULL) &&
							(pcon->argsq[c_argcstl[side][i].isqKingTo].ppi == NULL) &&
							(!FAttacked(pcon, c_argcstl[side][i].isqRookTo, side ^ 1))) {
                                addMove( pcm, isqFrom, c_argcstl[side][i].isqKingTo, 
							        cmkNONE, cmfCASTLE);
						}
		case pcKNIGHT:	// No "break" statement above.
			//
			//	Knight is a non-sliding piece.  The king also drops into this
			//	code when it gets done messing around with castling.
			//
			for (pvec = s_argpvecPiece[ppiFrom->pc]; *pvec; pvec++)
				if (!((isqTo = isqFrom + *pvec) & 0x88))
					if ((ppiTo = pcon->argsq[isqTo].ppi) == NULL) {
						if (!inQuiescent) {
                            addMove( pcm, isqFrom, isqTo, cmkNONE, cmfNONE);
						}
					} else if (ppiTo->co != side) {
                        addMove( pcm, isqFrom, isqTo, MVV_LVA[ppiFrom->pc][ppiTo->pc], cmfCAPTURE);
					}
			break;
		default:
			Assert(false);	// Corrupt data check.
		case pcBISHOP:
		case pcROOK:
		case pcQUEEN:
			//
			//	Sliding piece.  This is all there is to it.
			//
			for (pvec = s_argpvecPiece[ppiFrom->pc]; *pvec; pvec++)
				for (isqTo = isqFrom + *pvec; !(isqTo & 0x88); isqTo += *pvec)
					if ((ppiTo = pcon->argsq[isqTo].ppi) == NULL) {
						if (!inQuiescent) {
                            addMove( pcm, isqFrom, isqTo, cmkNONE, cmfNONE);
						}
					} else {
						if (ppiTo->co != side) {
                            addMove( pcm, isqFrom, isqTo, MVV_LVA[ppiFrom->pc][ppiTo->pc], cmfCAPTURE);
						}
						break;
					}
			break;
		}
	}
	(pste + 1)->pcmFirst = pcm;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This generates moves, then removes the illegal ones.  Illegal moves are
//	defined as moves that leave us in check.

//	The cases where castling crosses check, or castling out of check is
//	attempted, are taken care of by VGenMoves.

void VGenLegalMoves(PCON pcon, PSTE pste)
{
	VGenMoves(pcon, pste, false);
    int side = pste->coUs;
    //	Delete all moves that leave our king in check.
	//
	for (PCM pcm = pste->pcmFirst; pcm < (pste + 1)->pcmFirst;) {

		MakeMove(pcon, pste, pcm);
		bool fLeftInCheck = inCheck( pcon, side );
		UnmakeMove(pcon, pste, pcm);
		if (fLeftInCheck)
			*pcm = *--(pste + 1)->pcmFirst;
		else
			pcm++;
	}
}
#endif
