
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	Gerbil
//
//	Copyright (c) 2001, Bruce Moreland.  All rights reserved.
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	This file is part of the Gerbil chess program project.
//
//	Gerbil is free software; you can redistribute it and/or modify it under
//	the terms of the GNU General Public License as published by the Free
//	Software Foundation; either version 2 of the License, or (at your option)
//	any later version.
//
//	Gerbil is distributed in the hope that it will be useful, but WITHOUT ANY
//	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
//	FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
//	details.
//
//	You should have received a copy of the GNU General Public License along
//	with Gerbil; if not, write to the Free Software Foundation, Inc.,
//	59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#include "engine.hpp"
#include "gproto.hpp"
#include "hash.hpp"
#include "stats.hpp"
#include <stdlib.h>

#ifdef	_DEBUG
static char const s_aszModule[] = __FILE__;
#endif

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This module has two basic functions.  It manages the repetition hash
//	table, and it manages the main transposition hash table.

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	These functions affect the repetition-check hash table, which is a small
//	table designed to hold all the moves in the current line I'm thinking
//	about, plus all the moves that have already been played on the board.

//	The hash table must be messed with in first-in/last-out order.  If you
//	hash element A, then you hash element B, then you remove element A, it is
//	not guaranteed that you will be able to find element B anymore, which is
//	very bad.  You must remove element B before element A.  Alpha-beta works
//	in this order naturally, so no problem.

//	If you mess up and leave orphan elements in this table while doing a
//	search, you risk a crash at best, and hard to find day-long bugs at worst.

//	This table must be a power of two in size.

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	The hash table is composed of two parallel tables, each with different
//	replacement schemes.  It ends up being an array of these:

typedef	struct tagHASHD {
	HASH	hashDepth;			// "Different search or deeper depth" table.
//	HASH	hashAlways;			// "Always replace" table.
}	HASHD, * PHASHD, * RGHASHD;


//HASHK s_arghashkPc[pcMAX][		// Any array of keys for each piece of each
//	coMAX][csqMAX];				//  color on each square.  This will be used
								//  to create a position hash key, and to
								//  update it when the engine moves pieces
								//  around.
HASHK s_arghashkBB[16][csqMAX];
HASHK s_arghashkEnP[filMAX];	// Hash keys to be XOR'd in if an en-passant
								//  capture is possible.
HASHK s_arghashkCf[cfMAX];		// Hash keys to beXOR'd in based upon castling
								//  legality.

//	This checks to see if this hash key is already represented in the hash
//	table, which indicates that a repetition has taken place.


bool FRepSet(PCON pcon, HASHK hashk)
{
	unsigned ihashk = (unsigned)hashk & (chashkREP - 1);
	for (;;) {
		if (pcon->arghashkRep[ihashk] == hashk)
			return true;
		if (!pcon->arghashkRep[ihashk])
			return false;
		ihashk = (ihashk + 1) & (chashkREP - 1);
	}
}

//	This sticks a hash key in the table.

void VSetRep(PCON pcon, HASHK hashk)
{
	unsigned ihashk = (unsigned)hashk & (chashkREP - 1);
	for (;;) {
		if (!pcon->arghashkRep[ihashk]) {
			pcon->arghashkRep[ihashk] = hashk;
			return;
		}
		ihashk = (ihashk + 1) & (chashkREP - 1);
	}
}

//	This removes a hash key from the table.

void VClearRep(PCON pcon, HASHK hashk)
{
	unsigned ihashk = (unsigned)hashk & (chashkREP - 1);
	for (;;) {
		if (pcon->arghashkRep[ihashk] == hashk) {
			pcon->arghashkRep[ihashk] = 0;
			return;
		}
		Assert(pcon->arghashkRep[ihashk]);
		ihashk = (ihashk + 1) & (chashkREP - 1);
	}
}

//	This function is called before the search is started, in order to
//	initialize the repetition table and prime it with positions that have
//	occured in the game.
//
//	The function only inserts an element if a position has already occured
//	twice.  You could change this to insert every unique position, but this
//	would result in some behaviors I think are stupid.  Some people like to
//	do it that way, but I don't.
//
//	If you leave it my way, sometimes the program will repeat a position.
//	This has some negative consequences, like you stand a higher chance of
//	getting a 50-move draw.
//
//	I changed this for release 02.  When deciding whether to add a repetition
//	element, the program looks at full keys.  It does this in order to avoid
//	trying to seize a draw in the case where a position has already appeared
//	twice, but with different castling or en-passant flags.

void VSetRepHash(PCON pcon)
{
	memset(pcon->arghashkRep, 0, sizeof(pcon->arghashkRep));
	for (int i = 0; i <= pcon->gc.ccm; i++) {

		for (int j = i + 1; j <= pcon->gc.ccm; j++)
			if (pcon->gc.arghashkFull[i] == pcon->gc.arghashkFull[j])
				if (!FRepSet(pcon, pcon->gc.arghashk[i])) {
					VSetRep(pcon, pcon->gc.arghashk[i]);
					break;
				}
	}
}


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	These functions deal with the transposition hash table.  The transposition
//	hash table is very large, and is actually two tables.

//	The first table operates via a "replace always" scheme.  If you try to
//	insert an element into the table, it always is inserted.

//	The second table operates via a "replace if deeper or more recent search"
//	replacement scheme.  Every time the "think" function is called, a sequence
//	number is incremented.  If you try to insert an element into the table, it
//	is inserted if the sequence number now is different than the one
//	associated with the element in the table.  If the sequence numbers are the
//	same, an element will be inserted only if the depth associated with the
//	element you are trying to add is greater than the depth associated with
//	the element already here.

//	This two table scheme is commonly known, although I found out about it via
//	an email from Ken Thompson in 1994 or so.  His idea is that deep searches
//	can stick around for a while, while stuff that is very recent can also
//	stay around, even if it is shallow.

//	His system also allowed for repetitions to be detected through this
//	main hash table, but I find that using a smaller table specifically for
//	that results in less complexity and (hopefully) fewer bugs.

//	If you make changes to the transposition hash, you should test the program
//	with Fine #70:

//	8/k/3p4/p2P1p2/P2P1P2/8/8/K7 w - - 0 1

//	The best move is Kb1 (a1b1), which should be found with a decent positive
//	score after a couple of seconds, or perhaps less, depending upon how
//	efficient this program ends up getting.  If you've broken the
//	transposition hash system, you won't solve this problem in sensible time.

//	It is very easy to screw up the transposition table hash, and it's
//	possible that this implementation has bugs already.

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	The hash table.

static RGHASHD s_rghashd;
static unsigned	s_chashdMac = 0;

//	This is used to control overwrites in one of the hash tables, as is
//	documented elsewhere in this module.  Look for "sequence number".

static U8 s_seq = 0;

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This checks both hash tables, looking primarily for an element that
//	causes a cutoff.  If neither element can produce a cutoff, I'll check
//	to see if either of them has a "best" move associated with it.  If so,
//	I'll remember the element, so I can increase the candidate move key (cmk)
//	for the move after I generate moves.

//	A "best" move is indicated by the "isqFrom" field being something other
//	than "isqNIL".  If it is "isqNIL", the other two move fields ("isqTo" and
//	"pcTo") are undefined.

//	It's very possible to have two different hash elements for this position,
//	each with a different best move.  If both elements have the same move,
//	I'll only remember one of them.


//	Given a hash element, this checks to see if conditions are ripe for a
//	cutoff.  In order to produce a cutoff, the element has to record a
//	search that's deep enough.  If so, an "exact" element always returns
//	a cutoff.  An "alpha" or "beta" element only returns a cutoff if it can
//	be proven that the score associated with it would result in a cutoff
//	given the current alpha or beta.


bool ProbeHashValue(PCON pcon, PSTE pste, int * pval, int Alpha, int Beta, int depth)
{
	PHASHD phashd = &s_rghashd[pste->hashkPc & s_chashdMac];
	pste->phashDepth = NULL;
	pste->phashAlways = NULL;
    // TODO: I added the test on s_seq so that I can analyze on the console
    // without the test I probe hash positions from a previous analysis
    PHASH phash = &phashd->hashDepth;
	if (phash->hashk == pste->hashkPc && s_seq == phash->seq ) {
        stats.Hdepthfound++;
        pste->phashDepth = phash;
        // special draw value, stop searching
//        if( phash->val == 0 )
//            return true;
        if (pste->plyRem <= phash->plyRem)
            return true;
    }
	return false;
}
#pragma optimize( "", on )

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This tries to update both hash tables with the results of this search.

//	A fine point here is that I don't overwrite the "best" move in an element
//	with a null move if there was already a "best" move in this element, as
//	long as the hash key in here is the key for the current position.

//	Meaning, if I found that this position was >=50 at some time past, I might
//	have a "best" move in here, which is the move that generated the score of
//	>=50.  Later on if I come back with a higher window, and I find that the
//	move is <=75 (a fail low), I won't have a best move from that search,
//	because none of the moves worked.  I'll keep the old best move in the
//	element, because it's possible that it's still best, and in any case it's
//	a good guess.


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
static void setMove(PHASH phash, CM *pcm) {
        phash->m1 = U8((pcm->m) & 0xff);
        phash->m2 = U8((pcm->m >> 8) & 0xff);
        phash->m3 = U8((pcm->m >> 16) & 0xff);
}
void RecordHashValue(PCON pcon, PSTE pste, const CM &cm, int val, int Alpha, int Beta, hashBound boundType)
{
	PHASHD phashd = &s_rghashd[pste->hashkPc & s_chashdMac];
    int hashf = hashfNONE;
	//
	//	Adjust mate scores.  Mate scores are weird because they change
	//	depending upon where in the tree they are found.  When stored in the
	//	hash table, additional weirdness can result.  I solve this problem by
	//	converting any mate scores to bounds, then storing the bounds.
	//
	//	A few weird cases were mates that are failing low, and -mates that are
	//	failing high.  These are stored, but I don't store any bound
	//	information, so it's not possible to cut off on these later.
	//
    // If we store the mat value as is (say a mat in 10), this value can propagate
    // to the root, at some point the hash table will say that all moves lead
    // to a mate if we allways use the value from the probe
    // We store a mate value that is higher that what could be find in the position
    // so the program will return the value of his pvs rather than the value
    // from an a mate coming from another line

//    if( val >= Beta )
//        hashf |= hashfBETA;
//    else if( val <= Alpha )
//        hashf |= hashfALPHA;
//    else
//        hashf |= hashfALPHA | hashfBETA;    // exact value
	//
	//	Replace "replace if deeper or more recent search" element only if
	//	the appropriate criteria are met.
	//
    PHASH phash = &phashd->hashDepth;
//    if( hashf != boundType )
        hashf = boundType;

	if ((s_seq != phash->seq) || (phash->plyRem < pste->plyRem)) {
        stats.Hdepthinsert++;
        if( s_seq == phash->seq && phash->hashk != pste->hashkPc)
            stats.Hcollide++;
        if( s_seq != phash->seq )
            stats.Hused++;
		phash->val = short(val);
		if (cm.m != 0) {
//            setMove(phash, pcm);
            phash->m1 = U8((cm.m) & 0xff);
            phash->m2 = U8((cm.m >> 8) & 0xff);
            phash->m3 = U8((cm.m >> 16) & 0xff);
		} else if (phash->hashk != pste->hashkPc)
			phash->m1 = phash->m2 = phash->m3 = 0;
		phash->hashk = pste->hashkPc;
		phash->hashf = U8(hashf);
        // special draw value
//        if( val == 0 )
//            phash->hashf = hashfEXACT;
		phash->plyRem = U8(pste->plyRem); //depth;
		phash->seq = s_seq;
	}
}
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	I didn't want to export this global throughout the rest of the project.

void VIncrementHashSeq(void)
{
	s_seq++;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Zero all hash memory.

void VClearHashe(void)
{
    if( s_chashdMac )
	    memset(s_rghashd, 0, (s_chashdMac + 1) * sizeof(HASHD));
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	"rand()" produces 15-bit values (0..32767).  I want 64 bits.
#pragma optimize( "", off ) 

static U64 U64Rand(void)
{
	return (U64)rand() ^ ((U64)rand() << 15) ^ ((U64)rand() << 30);
}
#pragma optimize( "", on ) 

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Fills up the hash key table with random gibberish.  These will be XOR'd
//	together in order to create the semi-unique hash key for any given
//	position.  They need to be as random as possible otherwise there will be
//	too many hash collisions.

//	I've never cared about hash collisions.  If they are a bigger problem
//	than I think, you might want to make this more random.

void VInitHashk(void)
{
/*	for (int pc = pcPAWN; pc <= pcKING; pc++)
		for (int isq = 0; isq < csqMAX; isq++)
            if (FilFromIsq(isq) <= filH) {
				s_arghashkPc[pc][coWHITE][isq] = U64Rand();
				s_arghashkPc[pc][coBLACK][isq] = U64Rand();
            }*/
	for (int fil = filA; fil <= filH; fil++)
		s_arghashkEnP[fil] = U64Rand();
	for (int cf = 0; cf < cfMAX; cf++)
		s_arghashkCf[cf] = U64Rand();
	for (int pc = 0; pc <= 15; pc++)
        for (int isq = 0; isq < csqMAX; isq++) {
			s_arghashkBB[pc][isq] = U64Rand();
        }
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This inits the main transposition hash table.  As of now this function
//	uses a constant for the table size, but this could easily be changed to
//	allow a variable.

//	The table must be a power of two in size or things will break elsewhere,
//	because I don't use "key % size" to get the element, I use
//	"key & (size - 1)", which only works with a power of two.

bool FInitHashe(PCON pcon, int cbMaxHash)
{
    if( cbMaxHash <= 0 || cbMaxHash > 1024 )
        cbMaxHash = 32;
    cbMaxHash *= 1024*1024;
	int chashdMax = 1;
	for (;;) {
		if (chashdMax * 2 * (int)sizeof(HASHD) > cbMaxHash)
			break;
		chashdMax *= 2;
	}
	VPrSendComment("%d Kbytes main hash memory (%d entries)", chashdMax * sizeof(HASHD) / 1024, chashdMax);
	if ((s_rghashd = (RGHASHD) malloc(chashdMax * sizeof(HASHD))) == NULL) {
		VPrSendComment("Can't allocate hash memory: %d bytes",
			chashdMax * sizeof(HASHD));
		return false;
	}
	s_chashdMac = chashdMax - 1;
    stats.Hsize = chashdMax;
	VClearHashe();
	return true;
}


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
