
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	Gerbil
//
//	Copyright (c) 2001, Bruce Moreland.  All rights reserved.
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	This file is part of the Gerbil chess program project.
//
//	Gerbil is free software; you can redistribute it and/or modify it under
//	the terms of the GNU General Public License as published by the Free
//	Software Foundation; either version 2 of the License, or (at your option)
//	any later version.
//
//	Gerbil is distributed in the hope that it will be useful, but WITHOUT ANY
//	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
//	FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
//	details.
//
//	You should have received a copy of the GNU General Public License along
//	with Gerbil; if not, write to the Free Software Foundation, Inc.,
//	59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#include "engine.hpp"
#include "gproto.hpp"
#include "eval.hpp"
#include "moves.hpp"
#include "genmagic.hpp"

#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#ifdef	DEBUG
static char const s_aszModule[] = __FILE__;
#endif

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

static HASHPawn *s_rghashp;			// Pawn hash table.
static unsigned	s_chashpMac = 0;	// Number of elements in the table, minus one.  I
							//  do the minus one thing in order to avoid an
							//  inner-loop subtract, which is no big deal, but
							//  every little bit helps.

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#ifdef _DEBUG

//	I sat here and thought about how to write this routine, for about five
//	minutes.  I could have written something that would have required
//	modification any time the pawn structure stuff is messed with, but I
//	didn't like that idea, since the code I've used in the past for this
//	routine gets kind of annyong to change.

//	On the other hand, this program is supposed to be simple and direct.

//	Finally, I could not resist doing something I have never done -- write
//	something that works like "printf".

//	The format string is comprised of commands.  The commands are as follows:

//		digit ('0'..'9')		The next argument is an array of "digit"
//								bitmaps, to be displayed vertically, with
//								"*" for any set bit and "." for any zero bit.
//								The argument after this is a string pointing
//								to the column header.

//		'!'						The next argument is an array of two bitmaps,
//								which should be displayed as one bitmap with
//								"P" if a bit from the first map is set, and
//								"p" if a bit from the second map is set, and
//								"." if neither bit is set.  There is nothing
//								special if both maps are set, that comes up
//								"P".  The argument after this is a string
//								pointing to the column header.

//	The column header strings should be <= 8 characters long.

//	Here's an example call:

//		VDumpPawnMaps("!1", argbmLoc, "Loc", &bmDoubled, "Doubled");

//	This would be somewhat different, but would work:

//		VDumpPawnMaps("21", argbmLoc, "Loc", &bmDoubled, "Doubled");

//	There's an example line somewhere in the code below, unless I accidentally
//	deleted it.  You can comment it out and mess with it, if you decide you
//	want to add more pawn bitmaps and dump them.

//	I'm glad I wrote this routine, rather than just assuming that my code
//	worked.  It found a bunch of bugs.


#define Idx(rank, file)		(file | (rank << 3))

void VDumpPawnMaps(char * szFmt, ...)
{
	va_list	pvaArg;
	char * sz;

	//	Display headers and figure out how many bitmaps in the largest array.
	//
	int cbmMax = 0;
	va_start(pvaArg, szFmt);
	for (sz = szFmt; *sz != '\0'; sz++) {
		char *	szArg;
		bitboard *pbmArg;
		int	cbm;

		if (isdigit(*sz))
			cbm = *sz - '0';
		else {
			Assert(*sz == '!');
			cbm = 1;
		}
		if (cbm > cbmMax)
			cbmMax = cbm;					// <-- New largest.
		pbmArg = va_arg(pvaArg, bitboard *);		// Pull bitmap pointer.
		szArg = va_arg(pvaArg, char *);		// Pull column header.
		if (sz != szFmt)
			putchar(' ');
		printf("%-8s", szArg);				// Display arg.
	}
	putchar('\n');
	//
	//	Display the bitmaps.  The largest argument I got was "cbmMax", so I
	//	will loop that many times, displaying one row of bitmaps each time.
	//
	for (int ibm = 0; ibm < cbmMax; ibm++) {

		if (ibm != 0)
			putchar('\n');
		//
		//	Display a row of bitmaps.
		//
		for (int rnk = rnk8; rnk >= rnk1; rnk--) {
			//
			//	Display one rank from all the bitmaps.
			//
			va_start(pvaArg, szFmt);
			for (sz = szFmt; *sz != '\0'; sz++) {
				char *	szArg;

				if (sz != szFmt)
					putchar(' ');
				int cbm = (isdigit(*sz)) ? *sz - '0' : 1;
				bitboard *pbmArg = va_arg(pvaArg, bitboard *);
				szArg = va_arg(pvaArg, char *);
				if (ibm >= cbm)			// This sometimes leaves trailing
					printf("        ");	//  spaces, but it's not worth fixing.
				else		// Display one rank from one bitmap.
					for (int fil = filA; fil <= filH; fil++) {
						bitboard bm;

						int isq = Idx(rnk, fil);
						bm = (U64)1 << isq;
						if (*sz != '!')
							putchar((bm & pbmArg[ibm]) ? '*' : '.');
						else if (bm & pbmArg[coWHITE])
							putchar('P');
						else
							putchar((bm & pbmArg[coBLACK]) ? 'p' : '.');
					}
			}
			putchar('\n');
		}
	}
}
#endif

static const int pawnDir[coMAX] = {16,-16};

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Penalties for pawn defects.

#define	valDOUBLED_ISO	25	// Doubled isolated pawn.
#define	valDOUBLED		15	// Normal doubled pawn, not strictly isolated.
#define	valISOLATED		15	// Strictly isolated pawn (no pawns of this color
							//  on either adjacent file).
#define valISOLATED_OPEN 20 // isolated pawn on semi-open file

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This gets the pawn structure info from the hash table if it's there,
//	otherwise it generates it.

//	In most cases, the stuff is in the hash table.  The hash table is big
//	enough, and the pawn structures don't change that much.

//	Since the pawn computation stuff isn't done that much, I don't take
//	special pains to make sure that it's fast.  It's better to be clear and
//	accurate.  It would be very possible to generate the defect values via
//	less complicated data structures, and fewer passes over the data, but
//	the excess time shoudn't matter.

//	On the other hand, the whole thing stays flexible so it should be easy
//	to add to.

//	This routine returns values from the perspective of the side to move,
//	and stores values in the hash table from white's perspective.

//	There are a lot of other pawn features that can be detected, but doubled
//	and isolated pawns should be good enough for now.  The trick is to play
//	solidly enough that the program doesn't drop pawns a few moves over its
//	horizon.

//	An obvious addition is *anything* that would help with the passed pawn
//	problem, which is always severe in a simple program.  The program could
//	also benefit from an understanding of backwardness and artificial
//	isolation.  Additionally, the values given for doubled and isolated pawns
//	could vary depending upon where the pawns are on the board, and the
//	condition of the opponent's pawns (a doubled isolated pawn on an open file
//	is worse than one on a closed file).

//	In a heavier implementation, some of the pawn maps might end up being
//	stored in the hash table itself, which is another level of complication.

static void ValPawns(const CON *pcon, const STE *pste)
{
	HASHPawn *phashp = &s_rghashp[pste->hashkPn & s_chashpMac];

	int	argval[coMAX];				// Total value of pawn features.
	bitboard argbmLoc[coMAX];			// Bits set for pawns that exist.
	BM	bmDoubled;					// Bit set indicates doubled pawn.
	BM	bmIsolated;					// Bit set indicates isolated pawn.
    bitboard bmInfluence[coMAX] = {0,0};

    argbmLoc[W] = pcon->pieces[WP];
    argbmLoc[B] = pcon->pieces[BP];
    phashp->pawnon6or7 = false;

	bmDoubled.qw = 0;
	bmIsolated.qw = 0;
	for (int coUs = coBLACK; coUs <= coWHITE; coUs++) {
        phashp->side[coUs].c_doubled = 0;
        phashp->side[coUs].c_islands = 0;
        phashp->side[coUs].blocked   = 0;
        phashp->side[coUs].pawnatk   = 0;

        for(int f = 0; f < 8 ; f++)
            phashp->side[coUs].c_pawn[f] = 0;

		argval[coUs] = 0;
        // build the influence bitmap, squares that are attacked by pawns
        // on their position or if they move up/down
        // note that pawns can move from one file to another so this computation
        // can be wrong in some situations
	    for (bitboard pawns = pcon->pieces[PAWN|coUs]; pawns; ) {
		    int sqr = PopLSB(&pawns);
            phashp->side[coUs].pawnatk |= pawn_attacks[coUs][sqr];
            int rank = Rank(sqr);
            if( coUs == W ) {
                if( rank == 5 || rank == 6 )
                    phashp->pawnon6or7 = true;
                for(int sq = sqr; sq < 64 ; sq += 8 )
            	    bmInfluence[W] |= pawn_attacks[W][sq];
            } else {
                if( rank == 2 || rank == 1 )
                    phashp->pawnon6or7 = true;
                for(int sq = sqr; sq >= 0 ; sq -= 8 )
            	    bmInfluence[B] |= pawn_attacks[B][sq];
            }
            phashp->side[coUs].c_pawn[File(sqr)]++;
	        for (bitboard xpawns = pcon->pieces[PAWN|(coUs^1)]; xpawns; ) {
		        int xsqr = PopLSB(&xpawns);
                if( File(sqr) == File(xsqr) ) {
                    int d = Rank(sqr) - Rank(xsqr);
                    if(coUs == W)
                        d = -d;
                    if( d > 0 ) {
                        phashp->side[coUs].blocked |= ((U64)1) << sqr;
                        break;
                    }
                }
            }
        }

        phashp->side[coUs].location = argbmLoc[coUs];
		//	Collect doubled pawn info.  I'm see if there is more than one
		//	pawn of a certain color on the a-file, if there is I will OR
		//	all of those pawns into the "bmDoubled" bitmap, then I'll try
		//	the b-file, and so on.
		//
		bitboard bmI = (U64)0x0101010101010101;	// Start with a1...a8.
		for (;;) {
			//	Get all pawns of this color on this file.
			//
			bitboard bmS = argbmLoc[coUs] & bmI;
			//
			//	If more than one bit set in this map, all the pawns of
			//	this color on this file are doubled.
			//
			if (bmS & (bmS - 1))
				bmDoubled.qw |= bmS;
			//
			//	I'm going to make a map of all the pawns on files adjacent
			//	to this file.  If there are no pawns in that map, all the
			//	pawns on this file are isolated.
			//
			bitboard bmT = 0;
			if (!(bmI & 0x01))
				bmT |= bmI >> 1;
			if (!(bmI & 0x80))
				bmT |= bmI << 1;
			if (!(bmT & argbmLoc[coUs]))
				bmIsolated.qw |= bmS;
			//
			//	Set up to examine the next file.  0x80 is the h-file, so
			//	if we just examined it, bomb out of the loop, otherwise
			//	shift to the next file.
			//
			if (bmI & 0x80)
				break;
			bmI += bmI;	// Examine file one to the right.
		}
	}
    phashp->side[W].influence = bmInfluence[W];
    phashp->side[B].influence = bmInfluence[B];

	for (int coUs = coBLACK; coUs <= coWHITE; coUs++) {
        phashp->side[coUs].passed = phashp->side[coUs].location & 
            (~phashp->side[coUs].blocked) & (~phashp->side[coUs ^ 1].influence) ;
        // for each pawn of the board...
	    for (bitboard pawns = pcon->pieces[PAWN|coUs]; pawns; ) {
		    int sqr = PopLSB(&pawns);
			bitboard bmLoc;
			bmLoc = ((U64)1) << sqr;
			if (bmLoc & bmIsolated.qw) {
                int f = File(sqr);
                if( phashp->side[coUs^1].c_pawn[f] == 0 )
                    argval[coUs] -= valISOLATED_OPEN;
				if (bmLoc & bmDoubled.qw)
					argval[coUs] -= valDOUBLED_ISO;
				else
					argval[coUs] -= valISOLATED;
            } else if (bmLoc & bmDoubled.qw) {
					argval[coUs] -= valDOUBLED;
			}
		}
	}
#ifdef _DEBUG
    static int dump_count = 0;
    dump_count++;
    if( dump_count <= 1 ) {
        bitboard blocked[2] = {phashp->side[0].blocked,phashp->side[1].blocked};
        bitboard passed[2] = {phashp->side[0].passed, phashp->side[1].passed };
        VDumpPawnMaps("!111111",
            argbmLoc, "Loc",
            &bmDoubled.qw, "Doubled",
            &bmIsolated.qw, "Isolated",
            &phashp->side[W].influence, "WInfl.",
            &phashp->side[B].influence, "BInfl.",
//            &blocked, "blocked",
            &phashp->side[W].pawnatk, "WAtk.",
            &phashp->side[B].pawnatk, "BAtk."
        );
    }
#endif
    phashp->hashk = pste->hashkPn;
//	phashp->val = argval[coWHITE] - argval[coBLACK];
    phashp->side[W].val = argval[coWHITE];
    phashp->side[B].val = argval[coBLACK];

    // now we know if a pawn is
    // - isolated, doubled, blocked/candidate, passed
    // we also have the count of pawn per file
    // we can do rook_file_attack & location(xside) to know if a rook is on a semi open file
}

void EvalPawns(const CON *pcon, const STE *pste, PHASHPawn & phashp) {
	phashp = &s_rghashp[pste->hashkPn & s_chashpMac];

	if (phashp->hashk != pste->hashkPn) {
        ValPawns(pcon, pste);
    }
//	return (pste->side == coWHITE) ? phashp->val : -phashp->val;
}


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Zero all pawn hash memory.

void VClearHashp(void)
{
    if( s_chashpMac )
	    memset(s_rghashp, 0, (s_chashpMac + 1) * sizeof(HASHPawn));
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This allocates the pawn hash memory according to the max value passed in
//	by the user, then clears the table.

bool FInitHashp(PCON pcon, int cbMaxPawnHash)
{
    if( cbMaxPawnHash <= 0 || cbMaxPawnHash > 256*1024*1024 )
        cbMaxPawnHash = 1*1024*1024;

    int chashpMax = 1;
	for (;;) {
		if (chashpMax * 2 * (int)sizeof(HASHPawn) > cbMaxPawnHash)
			break;
		chashpMax *= 2;
	}
    int Hsize = chashpMax * sizeof(HASHPawn);
	VPrSendComment("%d Kbytes pawn hash memory", Hsize/1024);
    s_rghashp = (HASHPawn *) malloc(Hsize);
	if ( s_rghashp == NULL) {
		VPrSendComment("Can't allocate pawn hash memory: %d bytes", Hsize);
        chashpMax = 1024;
        Hsize = chashpMax * sizeof(HASHPawn);
	    VPrSendComment("%d Kbytes pawn hash memory", Hsize/1024);
        s_rghashp = (HASHPawn *) malloc(Hsize);
	    if ( s_rghashp == NULL) {
		    VPrSendComment("Can't allocate pawn hash memory: %d bytes", Hsize);
		    return false;
        }
	}
	s_chashpMac = chashpMax - 1;
	VClearHashp();
	return true;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
