
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	Gerbil
//
//	Copyright (c) 2001, Bruce Moreland.  All rights reserved.
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	This file is part of the Gerbil chess program project.
//
//	Gerbil is free software; you can redistribute it and/or modify it under
//	the terms of the GNU General Public License as published by the Free
//	Software Foundation; either version 2 of the License, or (at your option)
//	any later version.
//
//	Gerbil is distributed in the hope that it will be useful, but WITHOUT ANY
//	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
//	FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
//	details.
//
//	You should have received a copy of the GNU General Public License along
//	with Gerbil; if not, write to the Free Software Foundation, Inc.,
//	59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#include "engine.hpp"
#include "gproto.hpp"
#include "hash.hpp"
#include "board.hpp"
#include "moves.hpp"
#include "stats.hpp"
#include "search.hpp"
#include "eval.hpp"
#include <stdlib.h>

#ifdef	_DEBUG
static char const s_aszModule[] = __FILE__;
#endif

// enables the negascout search, undefine to have standard alpha beta search (for bug hunting)
// reduce the number of nodes
#define NULL_WINDOW

// reduce the number of nodes
#define NULL_MOVE

// allways extend depth by one ply when in check
// increase the number of nodes, find mates in less plies
//#define EXTEND_CHECK

// reduce the number of nodes
#define USE_HASH

// change move ordering => reduce the number of nodes
#define USE_KILLERS

#define QS_BEFORE_HASH      // note:probing hash first gives worse results

// MULTI_CUT is buggy here, I'll check that later
//#define MULTI_CUT

// late move reduction : ok it's not so bad even with my poor implementation
//#define USE_LMR // disabled to test things

// change move ordering : very bad results/random sorting
//#define HISTORY

// skip descending the tree for some useless xchange in the quiescent search
// reduce the number of nodes ?
#define PRUNE_QS

// ...
#define FUTILITY_PRUNING

// this will stop search when we have a score good enought (mate)
#define DISTANCE_PRUNING

#define RECAPTURE_EXTENSION


#define moveisCapture(pcm) (((pcm)->cmk & (cmkCAPTURE | cmkBADCAPT)) != 0)
#define moveisPromotion(pcm) ((Pc((pcm)->m)|1) == (B_PROM|1))
#define moveisPawnMove(pcm) ((Pc((pcm)->m)|1) == (PAWN|1))
#define mateValue(depth)    (-valMATE + (depth) - 1)
#define moveto7thRank(pcm)  ((Rank(To((pcm)->m)) == 6) || (Rank(To((pcm)->m)) == 1))


// the definition of the node types can be found for example in
// "Multi-Cut alpha/beta-pruning in game-tre search"

typedef enum _nodeType {
    nodePV,
    nodeCUT,
    nodeALL
} nodeType;

static int SearchQ(PCON pcon, PSTE pste, int Alpha, int Beta);
static int Search(PCON pcon, PSTE pste, int Alpha, int Beta, const nodeType nt, int extCount);


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

static void SortKillers(PSTE pste, int depth)
{
    U32 Killer1 = Killers[depth].Killer1.m;
    U32 Killer2 = Killers[depth].Killer2.m;
    foreachmove(pste, pcm)
        if ( (pcm->m == Killer1) || (pcm->m == Killer2) ) {
			pcm->cmk |= cmkKILLER;
            stats.Killers++;
		}
}

//	This will be used to figure out which move to search first.  It is very
//	important to try to search the best move first, since this will produce
//	a huge reduction in tree size.

//	Best moves are typically winning captures and moves that have been shown
//	to be good here in the past (PV move and hash table move).

static void PrepareSort(const STE *pste, int depth) {
    foreachmove(pste, pcm) {
        pcm->cmk = 0;
        if(Capt(pcm->m)) {
            pcm->cmk = MVV_LVA[Pc(pcm->m)][Capt(pcm->m)];
        }
        else if( moveisPawnMove(pcm) && moveto7thRank(pcm)) {
            // only checking the dest rank because we know it's a pawn
            pcm->cmk = (cmkKILLER | cmkCAPTURE) + 10000;
        } else {
            // set cmk to delta positional value
        }
    }
#ifdef HISTORY
//	for (PCM pcm = pste->pcmFirst; pcm < (pste + 1)->pcmFirst; pcm++) {
//        if( pcm->cmk == cmkNONE )
//            pcm->cmk = cmkNONE | History.get( pcm );
//    }
#endif
}

static void PrepareSortQS(const STE *pste, int depth) {
    foreachmove(pste, pcm) {
        U32 m = pcm->m;
        pcm->cmk = 0;
        if(Capt(m)) {
            pcm->cmk = MVV_LVA[Pc(m)][Capt(m)];
        }
    }
#ifdef HISTORY
//	for (PCM pcm = pste->pcmFirst; pcm < (pste + 1)->pcmFirst; pcm++) {
//        if( pcm->cmk == cmkNONE )
//            pcm->cmk = cmkNONE | History.get( pcm );
//    }
#endif
}

static inline void SortNext(const STE *pste, PCM pcm) {
	PCM	pcmBest = pcm;

	for (PCM pcmI = pcm + 1; pcmI < (pste + 1)->mlist_start ; pcmI++)
		if (pcmI->cmk > pcmBest->cmk)
			pcmBest = pcmI;
	CM cm = *pcm;
	*pcm = *pcmBest;
	*pcmBest = cm;
}

static inline int MaterialValue(const STE *pste) {
    return (pste->valPcUs + pste->valPnUs) - (pste->valPcThem + pste->valPnThem);
}

static inline int MaterialGain(const CON *pcon, const CM *pcm) {
    // Capt(move) == 0 => no capture, array return 0 in this case
//    return s_argvalPc[ Capt(pcm->m) ];
    return b_valPc[ Capt(pcm->m) ];
}

static inline int recapMaterialValue(const CON *pcon, const CM *pcm) {
    // Capt(move) == 0 => no capture, array return 0 in this case
    return exchangePieceValue[ Capt(pcm->m) ];
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This does a null-move search to see if the side not to move has a threat.
//	If they can't drive the score above alpha, even if they get to move twice
//	in a row, their positions probably sucks, and we'll prune this variation.
//	It's not called in endgames, because it has problems detecting zugzwang.

//  We don't need to do a null move if we know allready that there is threats around
//  In endgames any pawn move is a potential threat (score will vary a lot)

// Note: don't do null move in the PV

static int FThreat(PCON pcon, PSTE pste, int Alpha, int Beta, const nodeType nt, int extCount)
{
//	if ((pste->plyRem <= 0) || (pste->fNull))
//		return true;
	bNullMove(pcon, pste);
    // Reduced-depth search
    int R = 2;
    stats.NM2++;
    // we can reduce the depth further when there is lot of material on the board
#if 0   // this does not work, I get a lot more nodes
    if( (pste->plyRem >= 3) && (pste->valPcThem + pste->valPcUs > 7*valKNIGHT) ) {
        stats.NM3++;
        R = 3;
    }
#endif
    (pste + 1)->plyRem = pste->plyRem - 1 - R;
    int val = -Search(pcon, pste + 1, - Beta, - Beta + 1, nt, extCount);
	return val;
}


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Quiescent search.  This lets captures play out at the end of the search,
//	so I don't end up aborting a line at a stupid place.

//	The quiescent search doesn't see checks, so it will not detect checkmates,
//	and might actually end up making decisions based upon an illegal line.

//	This is one reason I don't record the PV while in quiescent search -- I
//	don't want people telling me the program is doing illegal stuff.

// 200 gives better results than nothing or 100 or 150 (on wac-141)
#define QS_MARGIN   200

static int SearchQ(PCON pcon, PSTE pste, int Alpha, int Beta)
{
	pcon->ss.nodes++;	// Increment global node counter (or not)
	pste->ccmPv = 0;
	bool fFound = false;
    int iaminCheck = pste->checkf;
    int  f_score = 0;
    bool f_prune = false;
    int  f_max   = valMAX;

//1. if SEE(move) <0 then prune 
//2. if cureval + mat(move) + delta < alpha then prune 
//3. if cureval + SEE(move) + bigger_delta < alpha then prune 

	//
	//	See if static eval will cause a cutoff or raise alpha.
	//
    // this will be the return value if king evasion does no capture
	int posVal = ValEval(pcon, pste, Alpha, Beta);
    int bestVal = posVal; // ? posVal : -2;
    if( posVal == 0 )
        posVal = posVal;

#ifdef PRUNE_QS
    if( !iaminCheck && ((pste->valPcUs + pste->valPcThem) > 2*valQUEEN) ) {
        f_score = posVal + QS_MARGIN ;
        if (f_score <= Alpha) {
            f_prune = true;
            f_max   = f_score;
        }
    }
#endif
    if( !iaminCheck ) {
	    if (bestVal >= Beta)
		    return bestVal;
        if (bestVal > Alpha)
            Alpha = bestVal;
    }
    // generate captures, if we are in check then generate all moves (evade)
	if (iaminCheck) {
		bGenEvasions(pcon, pste);
	}
	else {
		bGenCaptures(pcon, pste, pste->mlist_start);
//		bGenNonCaptures(pcon, pste, p_moves);
	}

    int depth = int(pste - pcon->argste) + 1;
    PrepareSortQS( pste, depth );
    //
	//	Iterate through capturing moves, to see if I can improve upon alpha
	//	(which may be the static eval).
	//
    foreachmove(pste, pcm) {

        SortNext(pste, pcm);
		bMove(pcon, pste, pcm->m);
        fFound = true;
        // we've generated all moves if in check, but we wont play them
        // we just look if we have a legal move
        if( iaminCheck && (Capt(pcm->m) == 0) ) {
		    bUndoMove(pcon, pste, pcm->m);
            continue;
        }
#ifdef PRUNE_QS
        if( fFound && (f_prune)  && (pste[1].checkf == 0)
            && !moveisPromotion(pcm) && !moveisPawnMove(pcm)
            && (f_max + MaterialGain(pcon, pcm) <= Alpha)
            ) {
                stats.FP_QS++;
    		    bUndoMove(pcon, pste, pcm->m);
                // check that
                int approx = f_max + MaterialGain(pcon, pcm);
                if( approx > bestVal ) {
                    // is that needed ?
                    bestVal = approx;
                }
                continue;
        }
#endif

		int val = -SearchQ(pcon, pste + 1, -Beta, -Alpha);
		bUndoMove(pcon, pste, pcm->m);
        if (val >= Beta) {
			return val;
        }
        if (val > bestVal) {
			bestVal = val;
            if (val > Alpha)
			    Alpha = val;
        }
	}
	if (!fFound) {
        // We have no move !!
		if (iaminCheck) {
            // Checkmate
			bestVal = mateValue(depth);
		}
            // perhaps draw but we did not gen all moves
	}
//    RecordHashValue(pcon, pste, pcmBest, bestVal, LOWERBOUND, Beta, depth);
	return bestVal;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Full-width search.  This is the heart of the "chess playing" part of this
//	program.  It performs a recursive alpha-beta search.

//#define LOWERBOUND  a
#define LOWERBOUND  Alpha

static int Search(PCON pcon, PSTE pste, int Alpha, int Beta, const nodeType nt, int extCount)
{
	int	val;
    pste->ccmPv = 0;
	CM cmBest(0,0);
    bool isPV = (nt == nodePV);

#ifdef QS_BEFORE_HASH
    if (pste->plyRem < 0) {
        int val = SearchQ(pcon, pste,  Alpha, Beta);
        return val;
    }
#endif

    int depth = int(pste - pcon->argste) + 1;
    if( depth >= (csteMAX-50) ) {
        return SearchQ(pcon, pste,  Alpha, Beta);
    }

	//	Increment global node counter, and figure out if it is time to call
	//	the interface callback, which is what allows for an interruptible
	//	search.  It's crucial that this be called now and then.
	//
	if (++pcon->ss.nodes >= pcon->ss.nodesNext) {
		//
		//	2000 nodes is maybe 1/50to 1/100 of a second.
		//
		pcon->ss.nodesNext = pcon->ss.nodes + 20000;
		VPrCallback(pcon);
		VCheckTime(pcon);
		if ((pcon->fAbort) || (pcon->fTimeout))
			return 0;
	}
	//
	//	I can return draws and cut off if I'm somewhere other than the
	//	first ply of depth.
	//
	//	Return draw score if this is a repeated node.
	//
	if (FRepSet(pcon, pste->hashkPc)) {
        stats.repCount++;
		return 0;
	}
	//	Check hash table, in order to get "best" moves, and try to cut
	//	off.  Don't cut off within two plies of the root, otherwise the
	//	program might allow stupid rep draws.
	//
#ifdef USE_HASH
//    if( nt == nodePV )
//        ProbeHashValue(pcon, pste, &val, Alpha, Beta, depth);
//	else 
    if ( (depth > 1) && ProbeHashValue(pcon, pste, &val, Alpha, Beta, depth) ) {
        PHASH phash = pste->phashDepth;
        int val = phash->val;
		switch(phash->hashf)
        {
			case hashfEXACT:
                stats.Hcutoff ++;
                return val;
                break;
			case hashfBETA:
				{
                    if( val >= Beta ) {
                        stats.Hcutoff ++;
						return val;
                    }
// this can not work as is
                    if( val > Alpha ) {
                        stats.Halpha ++;
//                        Alpha = val;
                    }

					break;
				}
			case hashfALPHA:
				{
                    if( val <= Alpha ) {
                        stats.Hcutoff ++;
						return val;
                    }
                    // the next lines are sometimes good, sometimes bad...
                    if( val < Beta) {
                        stats.Hbeta++;
//                        Beta = val;
                    }
					break;
				}
            default:
                Assert(0);
        }
	}
#endif

#ifndef QS_BEFORE_HASH
    if (pste->plyRem < 0)
        return SearchQ(pcon, pste,  Alpha, Beta);
#endif

	//
	// Check alpha/beta window against a mate score
    // Thx for the idea
	//
    int iaminCheck = pste->checkf;
#ifdef DISTANCE_PRUNING
    int i_am_mate_score = mateValue(depth + 1);
//    int i_am_mate_score = mateValue(depth+ ( iaminCheck ? 0 : 1) );
    if( i_am_mate_score > Alpha ) {
//        Alpha = i_am_mate_score;
        if(i_am_mate_score >= Beta)
            return Beta; //i_am_mate_score;
    }
    int i_will_mate_score =  -mateValue(depth + 1);
    if( i_will_mate_score < Beta ) {
//        Beta = i_will_mate_score + 2;
        if( i_will_mate_score < Alpha )
            return Alpha;
    }
#endif
	//	Mark this move so if I find it in a sub-tree I can detect a rep
	//	draw.
	//
	VSetRep(pcon, pste->hashkPc);

    int a = Alpha;
    int b = Beta;
    int bestVal = valMIN;  // a
    int preExtend = 0;
    int nextExtCount = extCount > 0 ? extCount - 1 : 0;

#ifdef NULL_MOVE
	//
	//	Do null-move search, which is a reduced depth that tries to
	//	figure out if the opponent has no threats.
	//
    if ((extCount == 0) && (!isPV) && (!iaminCheck) && (pste->valPcUs >= valROOK)
        && (pste->valPcThem >= valROOK)
       && (!pste->fNull)) {
//        PHASHPawn phashp;
//        EvalPawns(pcon, pste, phashp);
//        if( true || !phashp->pawnon6or7 ) {
        if( true  ) {
            int val = FThreat(pcon, pste, Alpha, Beta, nt, extCount);
            if( val >= Beta ) {
		        VClearRep(pcon, pste->hashkPc);
		        return Beta;
            }
            // extend one ply if there is a mate threat
            if( val <= -(valMATE-500) ) {
    #if 0
                char sz_score[50];
                sprintf(sz_score, "MATE threat in NM(%d)", val);
                VDumpLine(pcon, pste, sz_score);
                val = val;
    #endif
                preExtend = 1;
            }
        }
	}
#endif
	//	Generate moves.
	//
//    if( pste->hashkPc == 0x00001720c10998d8 ) {
//        Print(pcon, pste);
//        val = val;
//    }
	bGenMoves(pcon, pste);
	//
	//	I know how many reversible moves have been made up until now.
	//	I just tried to make another move.  If I am here, I didn't
	//	leave myself in check, so I succeeded.
	//
	//	If the number of reversible moves made before this is at
	//	least 50 for each side, this was a draw.
	//
    int lmc = legalMoveCount(pste);
	if ( (pste->plyFifty >= 50 * 2) && (lmc > 0) ) {
		pste->ccmPv = 0;
		VClearRep(pcon, pste->hashkPc);
		return 0;
	}
    // single move extension
    if( (lmc == 1)  )
        preExtend = 1;

    PrepareSort( pste, depth );
	//	Mark the PV and hash moves so they sort first.
	//
	if (lastPV[depth].m != 0) {
		SortPV(pste, depth);
		lastPV[depth].m = 0;
	}
#ifdef USE_HASH
	if (pste->phashDepth != NULL)
		SortHash(pste, pste->phashDepth);
#endif
#ifdef USE_KILLERS
    SortKillers(pste, depth);
#endif

#ifdef MULTI_CUT
    const R = 2;
    if( (nt == nodeCUT) && (!iaminCheck) && (pste->plyRem >= R) ) {
        int c = 0;
        int m = 0;
        const C = 3;
        const M = 10;
	    for (PCM pcm = pste->pcmFirst; pcm < (pste + 1)->pcmFirst && (m < M); pcm++) {
            SortNext(pste, pcm);
		    MakeMove(pcon, pste, pcm);
		    (pste + 1)->plyRem = pste->plyRem - 1 - R;
#ifdef EXTEND_CHECK
            if( (pste + 1)->fInCheck )
                (pste + 1)->plyRem++;
#endif
            int val = -Search(pcon, pste + 1, - b, - a, nodeALL, extCount);
    		UnmakeMove(pcon, pste, pcm);
            if( val >= Beta ) {
                c++;
                if( c >= C ) {
				    VClearRep(pcon, pste->hashkPc);
                    return Beta;
                }
            }
            m++;
        }
    }
#endif

#ifdef RECAPTURE_EXTENSION
    // margin to include Knight, Bishop and rook
    int prevCaptureValue    = pste->prevCaptValue;
    int prevCaptureValueMin = pste->prevCaptValue - 200;
    int prevCaptureValueMax = pste->prevCaptValue + 200;
    int prevCaptureSqr = pste->prevCaptSqr;
#endif

    bool fFound = false;
    hashBound hashf = hashfALPHA;
    hashBound hashfFound = (Beta == (Alpha + 1)) ?  hashfALPHA : hashfEXACT;
    int  c_move = 0;

    bool f_prune = false;
    int  f_max   = valMAX;
    int  f_score = 0;

// 300, 200 gives wrong results /
#define FUTILITY_MARGIN (300)
#define FRONTIER    0

#ifdef FUTILITY_PRUNING
    if( !iaminCheck && (pste->plyRem == FRONTIER) && ((pste->valPcUs + pste->valPcThem) >= 2*valQUEEN) ) {
//        int posVal = ValEval(pcon, pste, Alpha, Beta);
        f_score = MaterialValue( pste ) + FUTILITY_MARGIN ;
        if (f_score <= Alpha) {
            f_prune = true;
            f_max   = f_score;
        }
    }
#endif

	//
	//	Iterate through the move list, trying everything out.
	//
    foreachmove(pste,pcm) {
		//
		//	All moves are sorted on the cmk field
		//	cmk is set when generating moves or recomputed for hash/pv moves
		//
        SortNext(pste, pcm);
		//
		//	Make the move, then see if the 50-move counter 
        //  indicates that I need to return a draw score
		//
		bMove(pcon, pste, pcm->m);
        int extension = preExtend;
        // don't try to reduce if we just extended the search
        bool can_reduce = (extCount + extension) > 0 ? false : true;

        if( (pste + 1)->checkf ) {
            can_reduce = false;
#ifdef EXTEND_CHECK
            if( pste->plyRem == FRONTIER || extCount < 4) {
                extension = 1;
            }
#else
            // if I extend before max depth then I could get a doubled depth at the end
            // and it's a waste
            if( pste->plyRem == FRONTIER ) {
                extension = 1;
            }
#endif
        }
        // Pawn push to 7th
        if( moveisPawnMove(pcm) && moveto7thRank(pcm) ) {
            extension = 1;
            can_reduce = false;
        }
#ifdef RECAPTURE_EXTENSION
        if( prevCaptureValue && Capt(pcm->m) && (To(pcm->m) == prevCaptureSqr) ) { 
            int matValue = recapMaterialValue(pcon, pcm);
            if( (matValue >= prevCaptureValueMin) && (matValue <= prevCaptureValueMax)) {
                stats.recapture++;
                extension = 1;
                can_reduce = false;
            }
        }
#endif
        //  ==========================================================================
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
        if( extension && (extCount < 10) )
		    (pste + 1)->plyRem++;
        //
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        nodeType succNType;
        if(nt == nodeCUT )
            succNType = nodeALL;
        else if(nt == nodeALL || fFound)
            succNType = nodeCUT;
        else 
            succNType = nodePV;
        // the above does not work like the old code below, it gives the double of nodes ???
        // nodePV is only checked for null move so wth ?
// lets try a bit without the next line 
//        succNType = fFound ? nodeCUT : nodePV;

#ifdef FUTILITY_PRUNING

        if( fFound && (nt != nodePV) && (f_prune) && can_reduce
//        if( fFound && (f_prune) && can_reduce
            && !moveisPromotion(pcm) && !moveisCapture(pcm) && !moveisPawnMove(pcm)
//            && (f_max + MaterialGain(pcon, pcm) <= a)
            && (f_max  <= a)
            ) {
                stats.FP_frontier++;
    		    bUndoMove(pcon, pste, pcm->m);
                // check that
//                int approx = f_max + MaterialGain(pcon, pcm);
                int approx = f_max;
                if( approx > bestVal ) {
                    // is that needed ?
                    bestVal = approx;
			        cmBest = *pcm;
			        pste->argcmPv[0] = cmBest;
			        pste->ccmPv = 1;
                }
                continue;
            }
#endif

#ifdef USE_LMR
        // we still need to check history that's why i use a number as big as 20
        // note that without history this 20 can be too small
        // test on killer is redundant with move count
        if( (c_move >= 20) && (nt != nodePV) && (!iaminCheck) && can_reduce && (pste->plyRem >= 1) 
            && (pcm->cmf == cmfNONE)
            ) {
            // search with a reduced depth
            stats.LMR++;
		    (pste + 1)->plyRem --;
            val = -Search(pcon, pste + 1, - b, - a, succNType);
            if( val >= Beta ) {
//            if( val >= a ) {
                // re-search at normal depth
                stats.LMRreSearch++;
    		    (pste + 1)->plyRem ++;
                val = -Search(pcon, pste + 1, - b, - a, succNType);
            }
        }
        else
#endif
        val = -Search(pcon, pste + 1, - b, - a, succNType, nextExtCount + extension);

#ifdef NULL_WINDOW
#if 0
        // method 1 gives wrong result ?? ==> results ok now but i can't see the full pv
        if( (fFound) && (val > a) && (val < Beta) ) {
            stats.reSearch++;
           val = -Search(pcon, pste + 1, - Beta, - val, nodePV, nextExtCount + extension);
        } 
#else
// the second method gives a lot more of re-search but nearly the same number of nodes
        if( (fFound) && (val > a) && ( val < Beta) ) {
            val = -Search(pcon, pste + 1, - Beta, - a, nodePV, nextExtCount + extension);
            stats.reSearch++;
        }
#endif
        /*else
            if( (nt == nodePV) && (val == Beta) && (Beta == a + 1) ) {
                stats.reSearch++;
                val = a;
                val = -Search(pcon, pste + 1, - Beta, - val, succNType);
            }*/
#endif

		bUndoMove(pcon, pste, pcm->m);
		if ((pcon->fAbort) || (pcon->fTimeout)) {
			VClearRep(pcon, pste->hashkPc);
			return 0;
		}
		//	We got a value back.  We unmade the move.  We're not dead.  Let's
		//	see how good this move was.  If it was >= "beta", it was so good
		//	that we don't need to search for anything better, so we'll leave.
		//
		//	If it was not >= "beta", but it was > "alpha", this is better than
		//	anything else we've found before, but not so good we have to
		//	leave.  These kinds of moves are actually quite rare.  If I find
		//	one of these, I have to store it in the PV (main-line) that I'm
		//	constructing.  This might end up being the main-line for the whole
		//	search, if it gets backed up all the way to the root.
		//
        if( val > bestVal ) {
            bestVal = val;
		    if (val > a) {
			    if (val >= Beta) {
				    //
				    //	This move failed high
				    //
                    //  Set the new killer move only if it was not marked as good allready
                    //  'Good' moves are allready sorted first so storing them in the killer
                    //  array is conter productive
                    if( !moveisCapture(pcm) )
                        Killers[depth].set(pcm);
                    RecordHashValue(pcon, pste, *pcm, bestVal, LOWERBOUND, Beta, hashfBETA);
//                    if( nt == nodePV )
//                        History.update( pcm, depth );
				    VClearRep(pcon, pste->hashkPc);
				    return bestVal;
			    }
			    //	This move is between alpha and beta, which is actually pretty
			    //	rare.  If this happens I have to add a PV move, and append the
			    //	returned PV to it
			    //
                if( (pste + 1)->ccmPv > 0 )
			        memcpy(pste->argcmPv + 1, (pste + 1)->argcmPv,
				        (pste + 1)->ccmPv * sizeof(CM));
			    cmBest = *pcm;
			    pste->argcmPv[0] = cmBest;
			    pste->ccmPv = (pste + 1)->ccmPv + 1;
			    a = bestVal;
                hashf = hashfFound;
		    }
        }
		fFound = true;
        c_move++;

#ifdef NULL_WINDOW
        b = a + 1;              // set the new null window
#endif
	}
	VClearRep(pcon, pste->hashkPc);
	if (!fFound) {
        // We have no move !!
		if (iaminCheck) {
            // Checkmate
			bestVal = mateValue(depth);
        } else {
            // Draw
            bestVal = 0;
//            char sz_score[50];
//            sprintf(sz_score, "stalemate");
//            VDumpLine(pcon, pste, sz_score);
        }
        hashf = hashfFound;
	}
    RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, hashf);
//    if( nt == nodePV )
//        History.update( pcmBest, depth );
	return bestVal;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Full-width search.  This is the heart of the "chess playing" part of this
//	program.  It performs a recursive alpha-beta search.

static struct {
    char    move[12];
    int     c_nodes;
    int     hval;
} sortDebug[200];
static int c_sortDebug;

#define maxRootMoves    200
static struct _rootMove {
    char szmove[8];
    int score;
    int nodes;
    CM cm;
} rootMoves[maxRootMoves];
static int rootMovesCount;

int ValSearchRoot(PCON pcon, PSTE pste, int Alpha, int Beta)
{
    int a = Alpha;
    int b = Beta;
    int bestVal = valMIN;
    int	val;
    int depth = int(pste - pcon->argste + 1);
    c_sortDebug = 0;

    //	Increment global node counter
    //
    ++pcon->ss.nodes;

    //	Stuff is doone a little differently at the root than at any other
    //	ply.
    //
    //
    //	Mark this move so if I find it in a sub-tree I can detect a rep
    //	draw.  I don't check to see if a rep is already set, because if
    //	we are here, we have to be making a move, not whining about a
    //	blown chance for a draw.  I also don't try to cut off from the
    //	hash table, for the same reason.  I want a move, not a score.
    //
    //	I do probe the table though, in case there is "best move" info
    //	in there.
    //
    (void) ProbeHashValue(pcon, pste, &val, Alpha, Beta, depth);

    VSetRep(pcon, pste->hashkPc);
    //
    //	If it's the first time we come here (depth == 0)
    //	generate the moves
    //
    if( pste->plyRem == 0 ) {
        // first time with this position
        bGenMoves(pcon, pste);
        pcon->ss.ccmLegalMoves = legalMoveCount(pste);
        rootMovesCount = pcon->ss.ccmLegalMoves;
        foreachmove(pste,pcm) {
            CmToSz( pcm, rootMoves[iMove].szmove );
            rootMoves[iMove].cm     = *pcm;
            rootMoves[iMove].score  = - iMove; // keep the generation order
            if( Capt(pcm->m) )
                rootMoves[iMove].score -= 100;  // captures at the end
            rootMoves[iMove].nodes  = 0;
        }
    }
    {
        // after the first time (depth iteration)
        // sort the moves by node count
#ifdef USE_HASH
        PHASH phash = pste->phashDepth;
        U32 hm = phash ? hashmove(*phash) : 0;
#endif
        for(int i = 0; i < rootMovesCount; i++) {
            PCM pcm = &rootMoves[i].cm;
            if ( rootMoves[i].cm.m == lastPV[1].m )
                rootMoves[i].score += 100000;
#ifdef USE_HASH
		    if ( pcm->m == hm )
                rootMoves[i].score  += 10000;
#endif
        }
    }
	//	Mark the PV and hash moves so they sort first.
	//
    // bubble sort
    for(int i = 0; i < rootMovesCount - 1; i++) {

        for(int j = i+1 ; j < rootMovesCount; j++)
            // most moves have the same score because of null window search
            if( (rootMoves[j].score > rootMoves[i].score) ||
                (rootMoves[j].score == rootMoves[i].score && (rootMoves[j].nodes > rootMoves[i].nodes) )) {
                _rootMove swap = rootMoves[i];
                rootMoves[i] = rootMoves[j];
                rootMoves[j] = swap;

            }
    }

    CM cmBest(0,0);
	bool fFound = false;
    hashBound hashf = hashfALPHA;
	//
	//	Iterate through the move list, trying everything out.
	//
//	for (PCM pcm = pste->pcmFirst; pcm < (pste + 1)->pcmFirst; pcm++) {
    for(int i = 0; i < rootMovesCount; i++) {
        PCM pcm = &rootMoves[i].cm;
		//
		//
		//	Things are done differently at the root of the search.
		//
        SAN	san;
        //	Record information about which move I am searching ...
        //
        pcon->ss.icmSearching = i+1;
        CmToSan(pcon, pste, pcm, &san);
        SanToSz(&san, pcon->ss.aszSearching);
        rootMoves[i].nodes = (int)pcon->ss.nodes;

#ifdef _DEBUG
        {
            sortDebug[c_sortDebug].hval = 0;
            sortDebug[c_sortDebug].c_nodes = (int)pcon->ss.nodes;
            strcpy(sortDebug[c_sortDebug].move , pcon->ss.aszSearching);
            strcat(sortDebug[c_sortDebug].move , " ");
            if( pcm->cmk & cmkPV )
                strcat(sortDebug[c_sortDebug].move , "V");
            if( pcm->cmk & cmkHASH )
                strcat(sortDebug[c_sortDebug].move , "H");
            if( pcm->cmk & cmkKILLER )
                strcat(sortDebug[c_sortDebug].move , "K");
            if( pcm->cmk & cmkCAPTURE )
                strcat(sortDebug[c_sortDebug].move , "C");
            if( pcm->cmk & cmkBADCAPT )
                strcat(sortDebug[c_sortDebug].move , "b");
            if( (pcm->cmk & cmkNONE) == cmkNONE)
                sortDebug[c_sortDebug].hval = pcm->cmk & ~0xfe000000;

        }
#endif
        //
        //	... then go make the move.
        //
        bMove(pcon, pste, pcm->m);
        bool can_reduce = true;
        int extension = 0;
        //
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
#ifdef EXTEND_CHECK
        if( (pste + 1)->checkf ) {
            can_reduce = false;
            // if I extend before max depth then I could get a doubled depth at the end
            // and it's a waste
            if( pste->plyRem == FRONTIER ) {
                (pste + 1)->plyRem++;
                extension = 1;
            }
        }
#endif
		//
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        nodeType succNType;
        if(fFound)
            succNType = nodeCUT;
        else 
            succNType = nodePV;

        val = -Search(pcon, pste + 1, - b, - a, succNType, extension);

#ifdef NULL_WINDOW
        // method 1 gives wrong result ?? ==> results ok now but i can't see the full pv
#if 0
        if( (fFound) && (val > a) && (val < Beta) ) {
            stats.reSearch++;
            val = -Search(pcon, pste + 1, - Beta, - val, nodePV, extension);
        }
#else
// the second method gives a lot more of re-search but nearly the same number of nodes
        if( (fFound) && (val > a) && (val < Beta) ) {
            val = -Search(pcon, pste + 1, - Beta, - a, nodePV, extension);
            stats.reSearch++;
        }
#endif
        /* else
            if( (val == Beta) && (Beta == a + 1) ) {
                stats.reSearch++;
                val = a;
                val = -Search(pcon, pste + 1, - Beta, - val, succNType);
            }*/
#endif
        // update the next sort key
        rootMoves[i].score = val;
        rootMoves[i].nodes = int(pcon->ss.nodes) - rootMoves[i].nodes;

        // debugging stuff
        sortDebug[c_sortDebug].c_nodes = int(pcon->ss.nodes) - sortDebug[c_sortDebug].c_nodes;
        c_sortDebug++;

		bUndoMove(pcon, pste, pcm->m);
		if ((pcon->fAbort) || (pcon->fTimeout)) {
			VClearRep(pcon, pste->hashkPc);
			return 0;
		}
		//	We got a value back.  We unmade the move.  We're not dead.  Let's
		//	see how good this move was.  If it was >= "beta", it was so good
		//	that we don't need to search for anything better, so we'll leave.
		//
		//	If it was not >= "beta", but it was > "alpha", this is better than
		//	anything else we've found before, but not so good we have to
		//	leave.  These kinds of moves are actually quite rare.  If I find
		//	one of these, I have to store it in the PV (main-line) that I'm
		//	constructing.  This might end up being the main-line for the whole
		//	search, if it gets backed up all the way to the root.
		//
        if( val > bestVal ) {
            bestVal = val;
		    if (val > a) {
			    if (val >= Beta) {
				    //
				    //	This move failed high, so we are going to return beta,
				    //	but if we're sitting at the root of the search I will set
				    //	up a one-move PV (if this isn't already the move I'm
				    //	planning to make), so this move will be made if I run out
				    //	of time before resolving the fail-high.
				    //
                    // Note:this code can not be reached because alpha==-inf and beta==+inf
                    // We can only be there when doing aspiration window search
				    if (memcmp(pcm, pste->argcmPv, sizeof(CM))) {
					    pste->argcmPv[0] = *pcm;
					    pste->ccmPv = 1;
				    }
				    //	This function needs the root moves generated, which
				    //	they are.
				    //
//                    if( (pcm->cmk & cmkMASK) < cmkCAPTURE )
//                        Killers[depth].set(pcm);
                    RecordHashValue(pcon, pste, *pcm, bestVal, LOWERBOUND, Beta, hashfBETA);
                    Beta = val;
				    VDumpPv(pcon, pcon->ss.plyDepth, Beta, prsaFAIL_HIGH);
//                    History.update( pcm, depth );
				    VClearRep(pcon, pste->hashkPc);
				    return Beta;
			    }
			    //	This move is between alpha and beta, which is actually pretty
			    //	rare.  If this happens I have to add a PV move, and append the
			    //	returned PV to it, and if I'm at the root I'll send the PV to
			    //	the interface so it can display it.
			    //
                if( (pste + 1)->ccmPv > 0 )
			        memcpy(pste->argcmPv + 1, (pste + 1)->argcmPv,
				        (pste + 1)->ccmPv * sizeof(CM));
			    cmBest = *pcm;
			    pste->argcmPv[0] = cmBest;
			    pste->ccmPv = (pste + 1)->ccmPv + 1;
			    a = bestVal;
                hashf = hashfEXACT;
			    //
			    //	This function needs the root moves generated, and they
			    //	are.
			    //	
			    VDumpPv(pcon, pcon->ss.plyDepth, a, prsaNORMAL);
#if 1 //_DEBUG
                stats.report();
#endif
		    }
        }
#ifdef NULL_WINDOW
        b = a + 1;              // null window
#endif
		fFound = true;
	}
#if 0
    printf("# SORT was ");
    for(int i = 0; i < c_sortDebug && i < 10; i++) {
        if( sortDebug[i].c_nodes < 10000 )
            printf("%s (%d) ", sortDebug[i].move, sortDebug[i].c_nodes );
        else
            printf("%s (%dk) ", sortDebug[i].move, sortDebug[i].c_nodes / 1000 );
        if( sortDebug[i].hval > 0 )
            printf("%d ", sortDebug[i].hval );
    }
    printf("\n");
#endif
	VClearRep(pcon, pste->hashkPc);
	if (!fFound) {
        // We have no move !!
		if (pste->checkf) {
            // Checkmate
			bestVal = -valMATE + 0;
		} else
            // Draw
            bestVal = 0;
        hashf = hashfEXACT;
	}
    RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, hashf);
//    History.update( pcmBest, depth );
	return bestVal;
}
